import pandas as pd
from polzybackend.models import User, Company, UserToCompany, CompanyToCompany, Role, user_company_roles
from polzybackend.utils.auth_utils import generate_token
import uuid
from flask_sqlalchemy import SQLAlchemy
from polzybackend import create_app
from ast import literal_eval
import sys
import io
import os
import xlrd
import json
from config import Config


def print_args():
    print("""    This program is used to download excel file or upload data from excel file in a polzy.db
    To download data use command = python ExcelDBManager.py --download
    To upload data use command   = python ExcelDBManager.py --upload path/to/excelfile.xlsx""")


def uuidBytes2String(byte):
    # used to convery uuid bytes to string
    if not byte:
        return
    elif isinstance(byte, bytes):
        string = str(uuid.UUID(bytes=byte))
        return string
    return byte


def string2bool(string):
    if isinstance(string, str):
        if string.lower().strip() == "true" or string.lower().strip() == "1":
            return True
    elif isinstance(string, int) or isinstance(string, float):
        if string >= 1:
            return True
    return False


def attributes_to_boolean(js):
    # converts only string "true" or "false" in boolean, because we don't need to convert int to bool in attributes
    try:
        attribute_json = json.loads(js)
    except:
        print(
            "Attributes are not in correct json format. Please make sure to use correct syntax like double inverted "
            "commas(\"), commas(,), etc. If you are using boolean values in Attributes. Else it is fine."
        )
        return js

    for key in attribute_json:
        if str(attribute_json[key]).lower().strip() == "true":
            attribute_json[key] = True
        elif str(attribute_json[key]).lower().strip() == "false":
            attribute_json[key] = False
    return json.dumps(attribute_json, ensure_ascii=False)


def download_excel(fileNameAndPath="db2excel.xlsx"):
    dfs = {}
    # reading database with pandas and storing table_name:dataframe(converted table) as key:value
    dfs[User.__tablename__] = pd.read_sql_table(User.__tablename__, Config.SQLALCHEMY_DATABASE_URI)
    dfs[Company.__tablename__] = pd.read_sql_table(Company.__tablename__, Config.SQLALCHEMY_DATABASE_URI)
    dfs[UserToCompany.__tablename__] = pd.read_sql_table(UserToCompany.__tablename__, Config.SQLALCHEMY_DATABASE_URI)
    dfs[CompanyToCompany.__tablename__] = pd.read_sql_table(CompanyToCompany.__tablename__, Config.SQLALCHEMY_DATABASE_URI)
    dfs[Role.__tablename__] = pd.read_sql_table(Role.__tablename__, Config.SQLALCHEMY_DATABASE_URI)
    dfs[user_company_roles.name] = pd.read_sql_table(user_company_roles.name, Config.SQLALCHEMY_DATABASE_URI)

    # converting uuid bytes to string
    dfs[User.__tablename__]["id"] = dfs[User.__tablename__]["id"].apply(lambda x: uuidBytes2String(x))
    dfs[User.__tablename__]["company_id"] = dfs[User.__tablename__]["company_id"].apply(lambda x: uuidBytes2String(x))
    dfs[Company.__tablename__]["id"] = dfs[Company.__tablename__]["id"].apply(lambda x: uuidBytes2String(x))
    dfs[UserToCompany.__tablename__]["user_id"] = dfs[UserToCompany.__tablename__]["user_id"
                                                                        ].apply(lambda x: uuidBytes2String(x))
    dfs[UserToCompany.__tablename__]["company_id"] = dfs[UserToCompany.__tablename__]["company_id"
                                                                        ].apply(lambda x: uuidBytes2String(x))
    dfs[CompanyToCompany.__tablename__]["child_id"] = dfs[CompanyToCompany.__tablename__]["child_id"
                                                                        ].apply(lambda x: uuidBytes2String(x))
    dfs[CompanyToCompany.__tablename__]["parent_id"] = dfs[CompanyToCompany.__tablename__]["parent_id"
                                                                        ].apply(lambda x: uuidBytes2String(x))
    dfs[user_company_roles.name]["user_id"] = dfs[user_company_roles.name]["user_id"
                                                                        ].apply(lambda x: uuidBytes2String(x))
    dfs[user_company_roles.name]["company_id"] = dfs[user_company_roles.name]["company_id"
                                                                        ].apply(lambda x: uuidBytes2String(x))

    with pd.ExcelWriter(fileNameAndPath, engine='xlsxwriter')as writer:
        for df in dfs:  # writing excel sheets with table name as sheet name
            dfs[df].to_excel(writer, df, index=False)
            worksheet = writer.sheets[df]  # pull worksheet object
            for idx, col in enumerate(dfs[df]):  # loop through all columns
                series = dfs[df][col]
                max_len = max((
                    series.astype(str).map(len).max(),  # len of largest item
                    len(str(series.name))  # len of column name/header
                )) + 1  # adding a little extra space
                worksheet.set_column(idx, idx, max_len)  # set column width
    return os.path.abspath(fileNameAndPath)


class UploadExcel:
    def __init__(self, fileName):
        """
        This class is used to upload data in polzy.db from excel sheet
        :param fileName: Excel file name with complete path
        """
        self.fileName = fileName
        self.uuids = {}  # stored uuids string as key and object as value which is later used when same uuid string is supplied
        self.db = self.get_db()
        self.users = {}  # stores User object which is later used in updating UserToCompany table
        self.companies = {} # stores Company object which is later used in updating UserToCompany and CompanyToCompany Table
        self.dfs = self.get_sanitized_dfs()

    def get_db(self):
        app = create_app(Config)
        return SQLAlchemy(app)

    def handleUUID(self, uid):
        # returns uuid bytes from string or create a new one and return it
        if not uid:
            return
        if uid not in self.uuids:
            try:
                self.uuids[uid] = str(uuid.UUID(uid))  # if the input id string is correct then it will generate uuid object
            except ValueError:  # if it is incorrect then it stores it as key and new uuid object as value
                self.uuids[uid] = str(uuid.uuid4())
        return self.uuids[uid]

    def get_sanitized_dfs(self):
        dfs = {}

        # User table replacing nan values with 0
        try:
            user_table = pd.read_excel(self.fileName, sheet_name=User.__tablename__)
        except xlrd.biffh.XLRDError:
            print("Given file is not a valid excel(xlsx) file!")
            sys.exit()
        user_table.fillna(0, inplace=True)
        dfs[User.__tablename__] = user_table

        # Company Table replacing nan values with 0
        company_table = pd.read_excel(self.fileName, sheet_name=Company.__tablename__)
        company_table.fillna(0, inplace=True)
        dfs[Company.__tablename__] = company_table

        # UserToCompany Table replacing nan values with 0
        user2company_table = pd.read_excel(self.fileName, sheet_name=UserToCompany.__tablename__)
        user2company_table.fillna(0, inplace=True)
        dfs[UserToCompany.__tablename__] = user2company_table

        # CompanyToCompany Table replacing nan values with 0
        company2company_table = pd.read_excel(self.fileName, sheet_name=CompanyToCompany.__tablename__)
        company2company_table.fillna(0, inplace=True)
        dfs[CompanyToCompany.__tablename__] = company2company_table

        roles_table = pd.read_excel(self.fileName, sheet_name=Role.__tablename__)
        roles_table.fillna(0, inplace=True)
        dfs[Role.__tablename__] = roles_table

        user_company_roles_table = pd.read_excel(self.fileName, sheet_name=user_company_roles.name)
        user_company_roles_table.fillna(0, inplace=True)
        dfs[user_company_roles.name] = user_company_roles_table

        return dfs

    def remove_nan(self, data):
        # removes key which has 0 as value because we have replace nan from 0 in dataframe
        removeable = []
        for dt in data:
            if data[dt] == 0:
                removeable.append(dt)
        for rm in removeable:
            del data[rm]

    def get_user(self, uid):
        if uid in self.users:
            return self.users[uid]
        user = self.db.session.query(User).filter(User.id == uid).first()
        return user

    def get_company(self, uid):
        if uid in self.companies:
            return self.companies[uid]
        company = self.db.session.query(Company).filter(Company.id == uid).first()
        return company

    def upload_user(self):
        user_table = self.dfs[User.__tablename__]
        user_table["id"] = user_table["id"].apply(lambda x: self.handleUUID(x))  # converting string uuid to bytes
        user_table["company_id"] = user_table["company_id"].apply(lambda x: self.handleUUID(x))
        data_dict = user_table.to_dict('r')  # converts dataframe into dict 'r' parameter is to remove index key
        for data in data_dict:
            self.remove_nan(data)
            user = self.db.session.query(User).filter(User.email == data["email"]).first()
            if not user:  # if data is not already present in database then create a new row
                user = User(**data)
            else:
                for key in data:  # if data their than update the existing database
                    if getattr(user, key):
                        setattr(user, key, data[key])
            self.users[user.id] = user  # saving the User object which is later use in updating other table
            self.db.session.add(user)
        try:
            self.db.session.commit()
        except Exception as ex:
            print(f"Exception while committing changes in db: {ex}")
            self.db.session.rollback()

    def upload_company(self):
        company_table = self.dfs[Company.__tablename__]
        company_table["id"] = company_table["id"].apply(lambda x: self.handleUUID(x))
        data_dict = company_table.to_dict('r')
        for data in data_dict:
            self.remove_nan(data)
            company = self.db.session.query(Company).filter(Company.name == data["name"]).first()
            if data.get("attributes"):  # If attribute is boolean convert it in boolean
                attributes = attributes_to_boolean(data.get("attributes"))
                data["attributes"] = attributes
            if not company:
                company = Company(**data)
            else:
                for key in data:
                    if getattr(company, key):
                        setattr(company, key, data[key])
            self.companies[company.id] = company
            self.db.session.add(company)
        try:
            self.db.session.commit()
        except Exception as ex:
            print(f"Exception while committing changes in db: {ex}")
            self.db.session.rollback()

    def upload_role(self):
        roles_table = self.dfs[Role.__tablename__]
        data_dict = roles_table.to_dict('r')
        for data in data_dict:
            self.remove_nan(data)
            roles = self.db.session.query(Role).filter_by(id=data['id']).first()
            if not roles:
                is_supervisor = True if str(data['is_supervisor']).upper() in ["TRUE", "1"] else False
                roles = Role(id=data["id"], name=data["name"], is_supervisor=is_supervisor)
            else:
                roles.name = data["name"]
                roles.is_supervisor = string2bool(data.get('is_supervisor', 0))
            self.db.session.add(roles)
        try:
            self.db.session.commit()
        except Exception as ex:
            print(f"Exception while committing changes in db: {ex}")
            self.db.session.rollback()

    def upload_user2company(self):
        user_company_roles_dict = self.dfs["user_company_roles"].to_dict('r')

        def get_role_id(user_id, company_id):  # used to get role id for current user & company relation
            for user_company in user_company_roles_dict:
                if user_company["user_id"] == user_id and user_company["company_id"] == company_id:
                    return user_company["role_id"]

        def get_role_object(id):  # returns role object from role id
            return self.db.session.query(Role).filter_by(id=int(id.strip())).first()

        user2company_table = self.dfs[UserToCompany.__tablename__]
        user2company_table["role"] = user2company_table.apply(lambda x: get_role_id(x["user_id"], x["company_id"]), axis=1)
        user2company_table["user_id"] = user2company_table["user_id"].apply(lambda x: self.handleUUID(x))
        user2company_table["user"] = user2company_table["user_id"].apply(lambda x: self.get_user(x) )
        user2company_table["company_id"] = user2company_table["company_id"].apply(lambda x: self.handleUUID(x))
        user2company_table["company"] = user2company_table["company_id"].apply(lambda x: self.get_company(x))
        data_dict = user2company_table.to_dict('r')
        for data in data_dict:
            self.remove_nan(data)
            # here 2 filters are used as we have company and user id and this sequence will not repeat because it is
            # expected that a user and a company will have only single relation
            user2company = self.db.session.query(UserToCompany
                                            ).filter(UserToCompany.user_id == data["user_id"]
                                                ).filter(UserToCompany.company_id == data["company_id"]).first()
            if data.get("attributes"):  # If attribute is boolean convert it in boolean
                attributes = attributes_to_boolean(data.get("attributes"))
                data["attributes"] = attributes
            if not user2company:
                user2company = UserToCompany(
                    user_id=data["user_id"], company_id=data["company_id"],
                    attributes=data.get("attributes", ""),
                    roles=[get_role_object(role) for role in str(data.get("role", "")).split(",")]
                )
            else:
                user2company.user_id = data["user_id"]
                user2company.company_id = data["company_id"]
                user2company.attributes = data.get("attributes", "")
                user2company.roles = [get_role_object(role) for role in str(data.get("role", "")).split(",")]
            self.db.session.add(user2company)
        try:
            self.db.session.commit()
        except Exception as ex:
            print(f"Exception while committing changes in db: {ex}")
            self.db.session.rollback()

    def upload_company2company(self):
        company2company_table = self.dfs[CompanyToCompany.__tablename__]
        company2company_table["parent_id"] = company2company_table["parent_id"].apply(lambda x: self.handleUUID(x))
        company2company_table["parent"] = company2company_table["parent_id"].apply(lambda x: self.get_company(x))
        company2company_table["child_id"] = company2company_table["child_id"].apply(lambda x: self.handleUUID(x))
        company2company_table["child"] = company2company_table["child_id"].apply(lambda x: self.get_company(x))
        data_dict = company2company_table.to_dict('r')
        for data in data_dict:
            self.remove_nan(data)
            # same like user relation, company relation is also expected to be unique
            company2company = self.db.session.query(CompanyToCompany
                                                 ).filter(CompanyToCompany.parent_id == data["parent_id"]
                                                    ).filter(CompanyToCompany.child_id == data["child_id"]).first()
            if data.get("attributes"):  # If attribute is boolean convert it in boolean
                attributes = attributes_to_boolean(data.get("attributes"))
                data["attributes"] = attributes
            if not company2company:
                company2company = CompanyToCompany(**data)
            else:
                for key in data:
                    if getattr(company2company, key):
                        setattr(company2company, key, data[key])
            self.db.session.add(company2company)
        try:
            self.db.session.commit()
        except Exception as ex:
            print(f"Exception while committing changes in db: {ex}")
            self.db.session.rollback()

    def upload_data(self):
        # Main method to upload all data in db
        self.upload_user()
        self.upload_company()
        self.upload_role()
        self.upload_user2company()
        self.upload_company2company()


if __name__ == "__main__":
    if len(sys.argv) < 2:
        print_args()
    elif "download" in sys.argv[1]:
        download_excel()
        print("File name db2excel.xlsx downloaded.")
    elif "upload" in sys.argv[1] and len(sys.argv) > 2:
        file = sys.argv[2]
        if not os.path.isfile(file):
            print(f"{file} doesn't exist. Please check file name or path is correct!")
        else:
            UploadExcel(file).upload_data()
            print(f"{file} successful update database.")
    else:
        print_args()



from requests import Session
from dataclasses import dataclass
from polzyFunctions.GlobalSystemConstants import GlobalSystemConstants

globalSystemConstants = GlobalSystemConstants()


@dataclass
class BackendSystemConnectorClasses():
    backendDict = {}


class SessionBuffer(object):
    # This class must be a singleton class, but due to conflict in overriding methods of subclass we can't make this
    # base class as singleton. Hence, wherever this class is inherited the sub class should be a singleton class
    # See the sample code of a subclass below:
    #
    # class SessionBufferSubClass(SessionBuffer):
    #     __instance__ = None
    #
    #     def __init__(self):
    #         super().__init__()
    #         if SessionBuffer.__instance__ is None:
    #             SessionBuffer.__instance__ = self
    #         else:
    #             raise UserWarning("Session-Klasse schon initialisiert")
    #
    #     @staticmethod
    #     def getInstance():
    #         if not SessionBuffer.__instance__:
    #             SessionBuffer()
    #         return SessionBuffer.__instance__
    #
    # Make sure to have __instance__ as class attribute, condition in __init__ to assign self to __instance__ if not
    # their & a staticmethod getInstance which you will call where ever you need this class instance.

    def __init__(self):
        self.sessionDict = {}

    def getSession(self, stage, sapClient):
        return


class ConnectorMeta:
    """
    Meta class for connections. Needs to be implemented and subclassed in each installation
    """
    def __init__(self, *args, **kwargs):
        pass

    def connectionSetup(self) -> Session:
        pass

    def executeGetRequestToURL(self, url: str, headers=None):
        return


class ConnectionFactory:
    """
    Connection factory should reply with an instance of ConnectorMeta
    """
    availableStages = globalSystemConstants.getAllStages()

    def __init__(self, stage):
        return

    def getConnector(self, endpointSystem: str, *args, **kwargs) -> ConnectorMeta:
        """
        The parameters are needed to identify which backend connector should be returned. This is only needed when
        there are more than one backend systems. Otherwise you can simply always return the same
        ConnectorMeta-Implementation
        :param endpointSystem:
        :param args:
        :param kwargs:
        :return:
        """
        return



import os


class Singleton(type):
    _instances = {}
    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            if cls not in cls._instances:
                cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


def get_file_path(fileName, basepath=""):
    if basepath and os.path.isfile(os.path.join(basepath, fileName)):
        return os.path.join(basepath, fileName)
    elif os.path.isfile(os.path.join(os.getcwd(), fileName)):
        return os.path.join(os.getcwd(), fileName)
    else:
        print(f"{fileName} path not found!")
        return fileName


from polzyFunctions.Activities.ActivitiesDataClasses import InputFields, FieldVisibilityTypes, FieldDefinition
from polzyFunctions.GlobalConstants import GlobalConstants, logger
from polzybackend import messenger
from polzyFunctions.ConnectionFactory import ConnectionFactory
from polzyFunctions.ConfigurationEngine.ConfigurationProvider import lConfigurationProvider
from polzyFunctions.GamificationActivityRecorder import recordActivityDecorator
from polzyFunctions.AntragActivityRecorder import recordAntragDecorator
from polzyFunctions.Dataclasses.CommonFieldnames import CommonFieldnames
from datetime import datetime
import json
import xml.dom.minidom
from polzyFunctions.LogLevelUpdater import LogLevelUpdater
from requests import Session
from polzyFunctions.Activities.ActivitiesDataClasses import InputFields


class Activity:
    """
    This is the template class for Activities. It's not called directly - just a template.

    """

    # constants to define post execution behavior of activity
    POSTEXECUTION_DEFAULT = 'default' # default behavior: update parent view
    POSTEXECUTION_LINK = 'link' # open link
    POSTEXECUTION_ACTIVE = 'active' # activity stays active after execution
    POSTEXECUTION_CLOSE = 'close' # close activity

    # parent instance flag
    PARENT_POLICY = 1
    PARENT_ANTRAG = 2

    def __init__(self):
        self.name = None           # set by each activity while intiating
        self.description = None    # set by each activity while intiating
        self.activityFields = InputFields()
        self.payload = None
        self.payloadJSON = None
        self.icon = None
        self.session: Session
        self.backendSystem = None  # Must be set by each activity before the execution
        self.stage = None
        self.latestResultFromRemoteRequest = None
        self.encrypt = False
        self.connectionToBackend = None   # fasifu

        # run --> default value
        # close --> closes activity
        # upload --> shows upload dialog
        # In case of run activity and any other (except "close"), the front-end makes call
        # to execute activity with new key activity (activity: activity.name) in payload.
        # usage: [{"name": "run", "caption": "Abschließen"}]
        self.frontendActions = []   # Actions, that are possible in the frontend for this Activity

        # post execution behavior of activity -- could redefined in activity if needed
        self.postExecuteBehavior = self.POSTEXECUTION_DEFAULT

        self.configurationProvider = lConfigurationProvider

    def toJsonForPersistence(self):
        """
        Will provide JSON-structure of current activity instance for persistence layer.
        :return:
        """
        return {
            "name": self.name,
            "description": self.description,
            "activityFields": self.activityFields.toJSON(),
            "payload": self.payload,
            "payloadJSON": self.payloadJSON,
            "icon": self.icon,
            "stage": self.stage,
            "postExecuteBehavior": self.postExecuteBehavior,
            "actions": self.frontendActions
        }

    def loadFromJsonFromPersistence(self, dic):
        """
        After antrags-instance was loaded from database, this method is called to also load state from JSON.
        :param dic:
        :return:
        """
        logger.debug(f"Updating {self.name} Activity class from json values.")
        for key, value in dic.items():
            if key == "activityFields":  # loading Field data in InputFields
                new_dic = {js.get("name"): (js.get("value") or js.get("valueChosenOrEntered")) for js in dic[key]}
                self.updateFieldValues(new_dic)
            else:
                setattr(self, key, value)
        return self

    def announceMessage(self, message, duration=3000, level='default', horizontal='left', vertical='bottom'):
        """
        Please see docu in Dataclasses/Antrag.py in announceMessage method
        :param message:
        :param duration:
        :param level:
        :param horizontal:
        :param vertical:
        :return:
        """
        logger.debug(f"Announce message: {message}")
        msg = json.dumps({
            'text': message,
            'autoHideDuration': duration,
            'variant': level,
            'anchorOrigin': {
                'horizontal': horizontal,
                'vertical': vertical,
            }
        })
        messenger.announce(f"data: {msg}\n\n")

    @recordAntragDecorator
    @recordActivityDecorator
    def executeActivity(self) -> bool:
        """
        This method is called when frontend "Execute"-Button is clicked. Depending on self.postExecutionBehavior
        usually the activity would be closed in the frontend after this. If you want to stop that (e.g. because
        of an error in the data) change the postExecutionBehavior accordingly.
        :return:
        """
        return True

    def _executePostRequestWithPayloadToURL(self, operation=None, usePutInsteadOfPost=None, headers=None):
        """
        Implement any specific logic for Post-Requests to a backend system. URL is taken from self.url!
        :param operation: any value to distinguish this call from other calls in the log
        :param usePutInsteadOfPost: self-explanatory
        :param headers: self-explanatory
        :return:
        """
        pass

    def _executeGetRequestToURL(self, url: str, headers=None):
        """

        :param url: the url to call to
        :param headers: header sequence for the call
        :return: answer from the backend system
        """
        try:
            return self.connectionToBackend.executeGetRequestToURL(url=url, headers=headers)
        except Exception:
            self.announceMessage(f"Problem occured while connecting to service address {url}", duration=0)

    def updateFieldValues(self, updateValues):
        """
        Frontend calls this method e.g. when "inputTriggersComplexUpdate" is set. We'll receive all current fields and
        their values and deal with them accordingly.
        :param updateValues: [{<field>: <value>, {},...]
        :return: nothing
        """
        if not updateValues:
            return

        logger.debug(f'New Values: {updateValues}')
        for name, value in updateValues.items():
            if name == CommonFieldnames.addressDict.value:
                if isinstance(value, dict):
                    for k, v in value.items():
                        self._updateFieldValue(k, v)
                continue

            self._updateFieldValue(name, value)

    def _updateFieldValue(self, name, value, optional=None):
        """

        :param name: fieldname of the activtiy
        :param value: current field value
        :param optional: True if the field might not exist (and not log message is requested in case it can't be found)
        :return:
        """

        # set parent flag
        if hasattr(self, 'polizze'):
            parent_flag = self.PARENT_POLICY
        elif hasattr(self, 'antrag'):
            parent_flag = self.PARENT_ANTRAG
        else:
            logger.debug(f"Activity {self.name} is not assigned to any policy or antrag instance")
            return

        lUpdatedField = self.activityFields.getField(name=name)
        if not lUpdatedField:
            # policy cse
            if parent_flag == self.PARENT_POLICY:
                logger.debug(f"Field provided {name} with value {value}. But not in policy activity")

            # antrag case
            # Maybe field from Antrag, not from activity:
            lUpdatedField = self.antrag.Fields.getField(name=name)
            if not lUpdatedField:
                if not optional:
                    logger.error(f"Field provided {name} with value {value}. But not in activity nor in Antrag")
                else:
                    logger.debug(f"Field provided {name} with value {value}. But not in activity nor in Antrag. "
                                 f"'Optional' is set, so no problem.")
                return

        lUpdatedField.value = value
        try:
            if parent_flag == self.PARENT_ANTRAG:
                self.antrag._updateTechnicalFieldValuesAfterChange(lUpdatedField)
        except Exception as e:
            logger.exception(f"Error in _updateTechnicalFieldValuesAfterChange (or Methode doesn't exist). "
                             f"Class was {self.__class__.__name__}. Error: {e}. Field: {name}, "
                             f"Value: {value}")

    def checkAndUpdateInputFields(self, inputFieldsToUpdate):
        """
        Here we can set Input Field values from external source. Each implementation will
        deal with the result differently
        :return:
        """

        if isinstance(inputFieldsToUpdate, InputFields):
            self.activityFields = inputFieldsToUpdate
        elif isinstance(inputFieldsToUpdate, FieldDefinition):
            lExistingField = self.activityFields.getField(name=inputFieldsToUpdate.name)
            lExistingField.__dict__.update(inputFieldsToUpdate.__dict__)
        elif isinstance(inputFieldsToUpdate, list):   # I guess this will be a list of Fields or one field
            for field in inputFieldsToUpdate:
                lExistingField = self.activityFields.getField(name=field.name)
                # Copy all attributes from newly added field into this field
                lExistingField.__dict__.update(field.__dict__)

    def _replaceVariablesInPayload(self):
        """
        Search in the payload string for $(<variablename>) and replace with either self.polizze or
        self.inputFields
        :return:
        """

        expression = ""
        if self.payload:
            expression = self.payload
            self.payload = self._replaceVariable(expression)
        elif self.payloadJSON:
            self.payloadJSON = self._getExpressionFromJson(self.payloadJSON)
        if not expression:
            return

    def deriveDefaultsFromUserOrCompany(self):
        """
        Here is the place to derive any default values from either user or company (or both) during initializiation
        :return:
        """
        pass

    def _getExpressionFromJson(self, jsonData):
        """
        This method loop through every element of json data and replace the strings with prefix - postfix $( - )
        to its actual placeholder by calling method _replaceVariable
        :param jsonData:
        :return:
        """
        if type(jsonData) == list:
            lis = []
            for item in jsonData:  # loop through each element and call this method recursively
                lis.append(self._getExpressionFromJson(item))
            return lis
        elif type(jsonData) == dict:
            dic = {}
            for key, value in jsonData.items():  # loop through each data and call this method recursively
                dic[self._getExpressionFromJson(key)] = self._getExpressionFromJson(value)
            return dic
        elif type(jsonData) == str:  # if str then replace the placeholder and return the data, no recursion after this
            return self._replaceVariable(jsonData)
        else:
            return jsonData

    def _replaceVariable(self, expression):
        if not "$(" in expression:
            return expression

        try:
            polizze = self.polizze
        except AttributeError:
            # In einem Antrag haben wir keine Polizze. Deshalb voll OK!
            pass

        try:
            antrag = self.antrag
        except AttributeError:
            pass

        while "$(" in expression:
            if expression[0:2] == "$(":
                left_part = ""
            else:
                left_part = expression.split("$(")[0]

            center = expression[len(left_part) + 2:]
            center = center.split(")")[0]

            try:
                right_part = expression[len(left_part) + len(center) + 3:]
            except Exception:
                right_part = ""

            # Left = part of payload before $( starts
            # right_part = part of payload after $(variable) ends
            # center = the <variable>-part inside $(<variable>)

            # center might be polizze.polizzenNummer or self.wirksamkeitsdatum
            if "." in center:
                lClassForAttribute = locals().get(center.split(".")[0])
                center = center.split(".")[1]
                try:
                    centerValue = getattr(lClassForAttribute, center)
                except Exception as e:
                    if hasattr(self, "polizze"):
                        logger.critical(f"Center was {center}, lClassForAttribute was: {lClassForAttribute}, "
                                        f"getattr didn't work. Policy is {self.polizze}. "
                                        f"Expression is: {expression}. Locals were : {locals()}")
                    elif hasattr(self, "antrag"):
                        logger.critical(f"Center was {center}, lClassForAttribute was: {lClassForAttribute}, "
                                        f"getattr didn't work. Antrag is {self.antrag}. "
                                        f"Expression is: {expression}. Locals were : {locals()}")
                    raise ValueError("Error in _replaceVariable. Please check logs.")
            else:
                lClassForAttribute = self
                centerValue = None
                lFound = False
                try:
                    lField = lClassForAttribute.antrag.Fields.getField(name=center)
                    if lField.valueTech:
                        centerValue = lField.valueTech
                    else:
                        centerValue = lField.value
                    lFound = True
                except AttributeError:
                    pass

                if not centerValue and not lFound:
                    try:
                        lField = lClassForAttribute.activityFields.getField(name=center)
                        if lField.valueTech:
                            centerValue = lField.valueTech
                        else:
                            centerValue = lField.value
                    except AttributeError:
                        centerValue = getattr(lClassForAttribute, center)

            expression = "".join([left_part, str(centerValue), right_part])

        return expression

    def _saveXMLorJSONResultAsPrettyPrinted(self, resultToPrint, operation: str = None):
        """
        Used for debugging when we want to save requests/responses from XML or JSON to files.
        :param resultToPrint:
        :param operation:
        :return:
        """
        if not LogLevelUpdater().saveJson:
             return

        if not operation:
            operation = self.__class__.__name__

        try:
            lFileNameResult = f"{self.configurationProvider.basePath}/JsonFilesDelete/{self.polizze.polizzenNummer}_" \
                              f"{operation}_" \
                              f"{datetime.now().strftime(GlobalConstants.dateFormatLong)}_reply.json"
        except Exception:
            if hasattr(self, "antrag"):
                lFileNameResult = f"{self.configurationProvider.basePath}/JsonFilesDelete/" \
                                  f"Antrag_{self.antrag.sapClient}_" \
                                  f"{self.antrag.productName.replace('/', '_')}_" \
                                  f"{operation}_{datetime.now().strftime(GlobalConstants.dateFormatLong)}_reply.json"
            else:
                lFileNameResult = f"{self.configurationProvider.basePath}/JsonFilesDelete/" \
                                  f"UnknownRootObject_" \
                                  f"{operation}_{datetime.now().strftime(GlobalConstants.dateFormatLong)}_reply.json"

        if isinstance(resultToPrint, dict):
            try:
                with open(lFileNameResult, "w") as resultFile:
                    json.dump(resultToPrint, fp=resultFile, indent=4)
                logger.debug(f"File written: {lFileNameResult}")
                return
            except Exception as e:
                logger.critical(f"Fehler beim JSON-Dump {e}")
                return

        if not resultToPrint:
            logger.info("Should have printed result of call, but there was nothing in it.")
            return

        try:
            lDOMString = xml.dom.minidom.parseString(resultToPrint)
            with open(lFileNameResult, "w") as resultFile:
                resultFile.write(lDOMString.toprettyxml())
            logger.debug(f"File written: {lFileNameResult}")
            return
        except Exception as e:
            logger.debug(f"Error during XML-Dump: {e}, will fallback to text dump.")

        # If we can't write it as XML nor JSON, maybe we can save a text at least:
        try:
            with open(lFileNameResult, "w") as resultFile:
                resultFile.write(str(resultToPrint))
            logger.debug(f"File written {lFileNameResult}")
        except Exception as e:
            logger.critical(f"Error {e} during writing of interface file. Filename was {lFileNameResult}. "
                            f"resultToPrint was {resultToPrint}. Error as ")

    def _getConnectionFactoryForActivity(self):
        """
        self-explanatory
        :return:
        """
        lFactory = ConnectionFactory(stage=self.stage)
        self.connectionToBackend = lFactory.getConnector(endpointSystem=self.backendSystem,
                                                         sapClient=self.sapClient)
        self.session = self.connectionToBackend.connectionSetup()

    def _addFieldForActivity(self, field: FieldDefinition):
        """
        Add a new field to the field catalog of this activity
        :param field:
        :return:
        """
        if not self.activityFields.getField(name=field.name):
            self.activityFields.addField(field)
        else:
            self.checkAndUpdateInputFields(field)

    def createFieldcatalogForActivity(self):
        """
        This method is used to update inputFields for current Activity and then update value of those inputFields
        :return:
        # ManageFieldCatalog.addFields(self)                     ## To update inputFields for current Activity
        # self.checkAndUpdateInputFields(self.activityFields)    ## To update inputFields value
        """
        pass

    @staticmethod
    def formatDateFieldForDateSelector(datetimeInput: datetime) -> str:
        return datetimeInput.strftime(GlobalConstants.dateFormat)

    def __str__(self):
        return self.__class__.__name__

    def toJSON(self):
        """
        Translates the fields and fieldgroups of this activity to JSON for the frontend.
        :return:
        """
        frontendDict = {
            'name': self.name,
            'description': self.description,
            'icon': self.icon,
            'postExecution': self.postExecuteBehavior,
            'fields': self.parseToFrontendFieldsWithoutGroup(),
            'actions': self.frontendActions,
        }

        lFieldGroups = self.parseToFrontendFieldGroups()
        if lFieldGroups:
            frontendDict.update(lFieldGroups)
            frontendDict.update(self.parseToFrontendFieldGroupFields())

        logger.debug(f"Activity dict for class {self.__class__.__name__} is \n {frontendDict}")
        return frontendDict

    def parseToFrontendFieldsWithoutGroup(self) -> list:
        """
        Fields, which are not included in groups are collected here
        :return:
        """
        lList = []

        for feld in self.activityFields.getAllInputFields():
            if feld.isGroupField:
                continue
            if feld.group:
                continue
            lList.append(feld.toJSON())

        return lList

    def parseToFrontendFieldGroups(self) -> dict:
        """
        Loops through the list of Antrags fields and identifies group-fields (if any).
        If they exist, a list of group-fields is returnd.
        :return:
        """
        lList = []

        for feld in self.activityFields.getAllInputFields():
            if feld.isGroupField:
                countfields = [x.name for x in self.activityFields.getAllInputFields()
                               if x.group == feld.name and x.fieldVisibilityType != FieldVisibilityTypes.hidden]
                if countfields:
                    lList.append(feld)

        if lList:
            return {"field_groups": [x.toJSON() for x in lList]}

        return {}

    def parseToFrontendFieldGroupFields(self) -> dict:
        """
        Will create one dict for each field-group in the field-catalog
        lDict[<groupname>] = [list of FieldDefinition.toJSON()

        :return:
        """
        lReturnDict = {}
        for feld in self.activityFields.getAllInputFields():
            if feld.fieldVisibilityType == FieldVisibilityTypes.hidden:
                continue
            if feld.group:
                if lReturnDict.get(feld.group):
                    lReturnDict[feld.group].append(feld.toJSON())
                else:
                    lReturnDict[feld.group] = [feld.toJSON()]

        return lReturnDict
